#define isb() asm volatile ("mcr p15, #0, %[zero], c7, c5,  #4" : : [zero] "r" (0) )
#define dsb() asm volatile ("mcr p15, #0, %[zero], c7, c10, #4" : : [zero] "r" (0) )
#define dmb() asm volatile ("mcr p15, #0, %[zero], c7, c10, #5" : : [zero] "r" (0) )

#define invalidate_instruction_cache()	asm volatile ("mcr p15, #0, %[zero], c7, c5,  #0" : : [zero] "r" (0) )
#define flush_prefetch_buffer()			asm volatile ("mcr p15, #0, %[zero], c7, c5,  #4" : : [zero] "r" (0) )
#define flush_branch_target_cache()		asm volatile ("mcr p15, #0, %[zero], c7, c5,  #6" : : [zero] "r" (0) )

#define invalidate_data_cache()			asm volatile ("mcr p15, 0, %0, c7, c6,  0\nmcr p15, 0, %0, c7, c10, 4\n" : : "r" (0) : "memory")
#define clean_data_cache()				asm volatile ("mcr p15, 0, %0, c7, c10, 0\n" \
	

#define ARM_CONTROL_MMU						(1 << 0)
#define ARM_CONTROL_STRICT_ALIGNMENT		(1 << 1)
#define ARM_CONTROL_L1_CACHE				(1 << 2)
#define ARM_CONTROL_BRANCH_PREDICTION		(1 << 11)
#define ARM_CONTROL_L1_INSTRUCTION_CACHE 	(1 << 12)

#define MMU_MODE	(  ARM_CONTROL_MMU			\
			 | ARM_CONTROL_L1_CACHE			\
			 | ARM_CONTROL_L1_INSTRUCTION_CACHE	\
			 | ARM_CONTROL_BRANCH_PREDICTION)

			 
			 
			 
			 
			 
			 
			 

#define CYCLE_COUNTER_ENABLE       (1<<0)  // enable all three counters
#define COUNTER_ZERO               (1<<1)  // reset two count registers
#define CYCLE_COUNTER_RESET        (1<<2)  // reset cycle counter register
#define CYCLE_COUNTER_DIVIDE_64    (1<<3)  // cycle count divider (64)

#define PMNC(v)  asm volatile("mcr p15, 0, %0, c15, c12, 0" :: "r"(v))
#define CCNT(v)  asm volatile("mrc p15, 0, %0, c15, c12, 1" : "=r"(v))


// https://www.raspberrypi.org/forums/viewtopic.php?f=63&t=155830
static inline void init_perfcounters (int32_t do_reset, int32_t enable_divider)
{
  // in general enable all counters (including cycle counter)
  int32_t value = CYCLE_COUNTER_ENABLE;

  // peform reset:  
  if (do_reset)
  {
    value |= COUNTER_ZERO;     // reset all counters to zero.
    value |= CYCLE_COUNTER_RESET;     // reset cycle counter to zero.
  } 

  if (enable_divider)
    value |= CYCLE_COUNTER_DIVIDE_64;     // enable "by 64" divider for CCNT.
    
  PMNC(value);

/*  
  value |= 16;
  // program the performance-counter control-register:
  asm volatile ("MCR p15, 0, %0, c9, c12, 0\t\n" :: "r"(value));  

  // enable all counters:  
  asm volatile ("MCR p15, 0, %0, c9, c12, 1\t\n" :: "r"(0x8000000f));  

  // clear overflows:
  asm volatile ("MCR p15, 0, %0, c9, c12, 3\t\n" :: "r"(0x8000000f));
  */
}
// http://infocenter.arm.com/help/topic/com.arm.doc.ddi0301h/DDI0301H_arm1176jzfs_r0p7_trm.pdf#page=270
//[6] ECCUsed to enable and disable Cycle Counter interrupt reporting:0 = Disable interrupt, reset value1 = Enable interrupt.

static inline unsigned int get_cyclecount (void)
{
  unsigned int value;
  // Read CCNT Register
  CCNT(value);
  return value;
}

// waiter in arm cycles
// that means in nano seconds
static inline unsigned int waitUntil(unsigned int waiter)
{
  unsigned int value;
  // Read CCNT Register
  do
  {
    CCNT(value);
  } while (value < waiter);
  return value;
}


void mmu_disable(void);
char * getErrorText(int errNo);
