/* What if we use for "timing" of the vectors the
 * smallest possible timer interrupt?
 * How many Nano is that?
 * 
 * That way we could use "marked" timing...   to try...
 * 
 * 
 * */


/*
 TODO
 Reset detection -> eleminate T1 -> RAMP, do with timer of Pi!
 
problematic but actually *should* work (but doesn't in all cases...
*/




/*  Version 0.3
	Done by Malban for the PiTrex project.
	last edit: 08th of Jamuary 2020
 ... not written down yet
	
	
Version 0.2

changes from 0.1
 - macro "syntax"
 - nano delays added, delay is not manually calibrated anymore
 - functions to access PSG registers
 - optimization for same VIA register states corrected (was using two different scales)
 - T1 timer differentiates whether T1 timer is used as 16 bit (very large scale) or usual 8bit
   (one access to IO port less)
 - read vectrex IO (joystick/buttons) only ONCE per wait recal call Possible
   (that also means you have to call waitrecal - or reset "ioDone" manually to zero before read)
 - probably some minor delay changes

	Possible enhancmence:
  - optimize scaling for very similar scale factors to use the same scale factor!
    (see optimization)
  - allow manually set of "no buzz vectrex" (calibration)
  - allow manually set of "cranky vectrex" (less cycles for Y integrator delay)
  - possibly allow manually settings of "draw lines in one go"
    since it is probably best, if different for each piTrex emulator
	- add float functions (draw move)
	- this would allow for easier transformations (screen scale, translocation, rotation etc...)
 
	WAIT STATES
  	Are implemented using a busy loop "v_delayCycles(int cycles)" (vectrex cycles).
    Values of the delaya are in nano second and should be "666" for a vectrex cycles.
    On bare metal that is true. On raspbian seemingly some other optimization kicks in, which SET_OPTIMAL_SCALE
    increases the speed of even a hardcoded assembler loop.
    In non bare metal the waiter seems to need about 940 "fake" nano seconds for one vectrex cycle.

  Optimization
   If not optimized, every vector and every positioning is drawn/done from zero. So every drawing is done with zeroed coordinates. This is SLOW!
   If OPTIMIZED than:
   - vectors "near" each other and with a similar same scale are drawn/positioned directly after another without zeroing

   As of now, each vector might be drawn using a different scale, since scale is directly calculated from the given
   16bit coordinate, so that the strength of the vector is optimized.
   On the one hand this is "good" if there are large differences in scale.
   If the scale difference is very subtle - than it would be more performat just to use the same scale.
   This kind of optimization is on the TODO list!
   (also possibly better looking - if scales are clustered, perhaps zero when scale changes)

*/

/***********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h> // these are fresstanding includes!
#include <stdint.h> // also "available":  <float.h>, <iso646.h>, <limits.h>, <stdarg.h>

#include <pitrex/pitrexio-gpio.h>
#include <pitrex/bcm2835.h>
#include "vectrexInterface.h"
#include "baremetalUtil.h"
#include "vectorFont.i" // includes font definition and string printing routines
#include "rasterFont.i" // includes font definition and string printing routines

/***********************************************************************/

/* values that we remember for possible optimization purposes */

int32_t currentCursorX; // 16 bit positioning value X (stored in long)
int32_t currentCursorY; // 16 bit positioning value Y (stored in long)
uint16_t currentScale; // currently active scale factor (T1 timer in VIA)
uint16_t lastScale; // last active scale factor (T1 timer in VIA)

uint8_t currentButtonState; // bit values as in vectrex
int8_t currentJoy1X; // -127 left, 0 center, +127 right
int8_t currentJoy1Y; // -127 up, 0 center, +127 down
int8_t currentJoy2X; // not used yet
int8_t currentJoy2Y; // not used yet
int8_t isCranky; // boolean: cranky should be checked during calibration! In "VecFever" terms cranky off = burst modus

int16_t consecutiveDraws; // how many lines/moves were drawn/made directly after another! (without zeroing)

unsigned int Vec_Rfrsh; // 30000 cylces (vectrex) = $7530, little endian = $3075

// these are 16 bit to allow illegal values!
// illegal values = unkown state
int16_t currentPortA; // == portA (also X SH) = current X Strength
int16_t currentYSH; // Y SH = current Y Strength
int16_t currentZSH; // Z SH = Brightness (if bit 7 is set -> brightness is off!)

int inCalibration;
int optimizationON;
int v_debug; // not actively used
uint8_t calibrationValue; // tut calibration

float sizeX;
float sizeY;
int16_t offsetX;
int16_t offsetY;

int calibrationStep;
int selectedMenu;
int selectionMade;
uint8_t ioDone; // vectrex IO should only be done once per "round"

uint32_t timerMark;

unsigned int psgShadow[16];
uint16_t ymLength;
uint16_t ymPos;
unsigned char *ymBuffer;
int ymloop;


/***********************************************************************/

/* should be called once on startup / reset
*/
void v_init()
{
	for (int i=0;i<16;i++)
		psgShadow[16] = 0;

  ioDone = 0;
  timerMark = 0;
  calibrationValue = 0;
  optimizationON = 1;
	inCalibration = 0;
	currentCursorX = 0; // 16 bit positioning value X
	currentCursorY = 0; // 16 bit positioning value Y
	currentButtonState = 0; // nothing pressed
	currentJoy1X = 0; // centered
	currentJoy1Y = 0; // centered
	currentJoy2X = 0; // centered
	currentJoy2Y = 0; // centered

	currentScale = 0; // currently active scale factor (T1 timer in VIA)
  lastScale = 1; // last active scale factor (T1 timer in VIA)
	consecutiveDraws = 0; // how many lines/moves were made after another!
	v_debug = 0;
	sizeX = 1.0;
	sizeY = 1.0;
	offsetX = 0;
	offsetY = 0;

	calibrationStep = 0;
	selectedMenu = 0;
	selectionMade = 0;


	// these are 16 bit to allow illegal values!
	currentPortA = 0x100; // == portA (also X SH)
	currentYSH = 0x100; // Y SH
	currentZSH = 0x100; // Z SH = brightness
	vectrexwrite (VIA_DDR_b, 0x9F); // All Outputs VIA Port B except COMPARE input and PB6 at Cartridge Port
	vectrexwrite (VIA_DDR_a, 0xFF); // All Outputs VIA Port A
#ifdef BEAM_LIGHT_BY_SHIFT
	SET (VIA_aux_cntl, 0x98); //Shift Reg. Enabled, T1 PB7 Enabled
#endif
#ifdef BEAM_LIGHT_BY_CNTL
	SET (VIA_aux_cntl, 0x80); // Shift reg mode = 000 free disable, T1 PB7 enabled 
#endif
	
	Vec_Rfrsh = 0x3075;//0x3075;//12405; // 30000 = $7530 little endian = $3075 = 12405
	SETW (VIA_t2, Vec_Rfrsh); //Set T2 timer to 30000 cycles and start timer and reset IFlag

	v_readButtons(); // read buttons to check if we should enter calibration on init

	// calibrate if button 1 is pressed on startup
	if ((currentButtonState&0x08) == (0x08)) inCalibration=1;


	ymLength = 0;
	*ymBuffer = 0;
	ymloop = 1;

	init_perfcounters(1, 0);
	
	unsigned int oCount = get_cyclecount();
	unsigned int difcount1 = get_cyclecount()-oCount;
	unsigned int difcount2 = get_cyclecount()-difcount1;
	unsigned int difcount3 = get_cyclecount()-difcount2;
	unsigned int difcount4 = get_cyclecount()-difcount3;
	unsigned int difcount5 = get_cyclecount()-difcount4;

//	
	printf("Cycle Counter dif 1: %i\r\n", difcount1);
	printf("Cycle Counter dif 2: %i\r\n", difcount2);
	printf("Cycle Counter dif 3: %i\r\n", difcount3);
	printf("Cycle Counter dif 4: %i\r\n", difcount4);
	printf("Cycle Counter dif 5: %i\r\n", difcount5);
}

/***********************************************************************/


/*
System timer runs at 250 Mhz
-> 1 tick = 1/250000000

0,000000004s
0,000004 ms
0,004 us
4 ns

1s
0,1 = 1/10
0,01 = 1/100
0,001 = 1 milli
0,000001 = 1 micro s
0,000000001 = 1 nano s


// 1 cycle = 1/1000000000
0,000000001s
0,000001 ms
0,001 us
1 ns
*/

// resolution "about" 4 nanoseconds
// but not less than OFFSET_CYCLE_OVERHEAD
void delayNano(uint32_t n)
{
    uint32_t nWait;
    nWait = n-OFFSET_CYCLE_OVERHEAD;
    if (nWait>((uint32_t)-20)) return;
/*
    if (n<1024)
    {
	nWait = n-OFFSET_CYCLE_OVERHEAD;
	if (nWait<=0) return;
    }
    else
    {
        nWait = n-OFFSET_TIMER_OVERHEAD;
        if (nWait>1024)
        {
            WAIT_MICRO_TIME(nWait>>10);
            nWait = nWait&0x3ff;
        }
    }
*/
    WAIT_CYCLE_NANO(nWait);
}

// delay in vectrex cycles
void v_delayCycles(uint32_t n)
{
    uint32_t nWait;
    nWait = (n*DELAY_PI_CYCLE_EQUIVALENT)-(OFFSET_CYCLE_OVERHEAD-3);
    if (nWait>((uint32_t)-20)) return;
    WAIT_CYCLE_NANO(nWait);
}
void v_delayCyclesQuarter()
{
    uint32_t nWait;
    nWait = (DELAY_PI_CYCLE_EQUIVALENT/4)-(OFFSET_CYCLE_OVERHEAD-3);
    if (nWait>((uint32_t)-20)) return;
    WAIT_CYCLE_NANO(nWait);
}

void v_delayCyclesEighth()
{
    uint32_t nWait;
    nWait = (DELAY_PI_CYCLE_EQUIVALENT/8)-(OFFSET_CYCLE_OVERHEAD-3);
    if (nWait>((uint32_t)-20)) return;
    WAIT_CYCLE_NANO(nWait);
}
/***********************************************************************/

/* Set current brightness of the Vectorbeam (provided BLANK is inactive).
   If bit 7 is set, vectrex does not show any brightness!
   It is checked if brightness is already set to the
   given value, if so - the function returns immediately.

   Ends without a delay, last set: VIA port B!
 */
void v_setBrightness(uint8_t brightness)
{
	if (brightness == currentZSH) return; // beware -> this might cause brightness drift to zero on completely same intensities all the time!
	if (brightness != currentPortA)
	{
		SET(VIA_port_a, brightness);
		currentPortA = (int8_t) brightness;
	}
	SET(VIA_port_b, 0x04); // MUX to intensity
	currentZSH = brightness;
	DELAY_ZSH();
	SET(VIA_port_b, 0x01); // SET port b to no Mux
}

/***********************************************************************/

/* Moves from the current "cursor" position to given offsets.
   Offsets are given in 16bit and are downscaled by
   "optimal" scale factors to vectrex strength/scale values.
   Wait is done using the Vectrex Timer T1.
   The wait is not "finished" - so other "pi" things can be done in between.
   After these "other pi things" are done "v_directDeltaMoveEnd()" must be
   called to insure T1 timer has expired.
 */
void v_directDeltaMove32start(int32_t _xLen, int32_t _yLen)
{
	int32_t xLen = _xLen*sizeX;
	int32_t yLen = _yLen*sizeY;
	SET_OPTIMAL_SCALE(xLen, yLen);
	SET_YSH16(yLen);
	SET_XSH16(xLen);
	START_T1_TIMER();
	currentCursorX += _xLen;
	currentCursorY += _yLen;
}
/*
Moves to the given position "absolut".
at the beginning the vectrex beam is rest to zero (center of screen).

The difference between a sequence of 
    v_zeroWait();
	v_directDeltaMove32start(_xLen, _yLen);

and a call to v_directMove32(...) is, that 
the absolut position call respects (can respect!) the 
chosen offsetX/offsetY.
Delta move do not (can not) respect the offsets!
*/
void v_directMove32(int32_t xEnd, int32_t yEnd)
{
    v_zeroWait();
    int32_t x = xEnd*sizeX+offsetX;
	int32_t y = yEnd*sizeY+offsetY;

	UNZERO(); // ensure vector beam can be moves
	SET_OPTIMAL_SCALE(x, y);
	SET_YSH16(y);
	SET_XSH16(x);
	START_T1_TIMER();
	consecutiveDraws=0;
	currentCursorX = xEnd;
	currentCursorY = yEnd;
	WAIT_T1_END();
}

/***********************************************************************/

/*
  Waits till T1 interrupt flag of vectrex VIA is set.
  Instead the Macro: WAIT_T1_END can be used!
  See also:
		v_directDeltaMove32start(int32_t xLen, int32_t yLen)
 */
void v_directDeltaMoveEnd()
{
	WAIT_T1_END();
	DELAY_CYCLES(1); // TEST: Gyrocks needs one additional cycle - seemingly some other VIA register is other set to soon. (Hershey Fonts))
}

/***********************************************************************/

/*	Draws "directly" (not buffered) to vectrex screen.
	All given coordinates are absolut screen coordinates.
 	Assume coordinates are in the range: -32768 - +32767,
	assuming 0,0 is in the middle of the screen.
 	"32" name tag denotes signed long range.

	This function acts "smart", it investigates given data (in regards to
    	last used data) and determines:
	- whether a move/zeroing is necessary
	- translates data to "delta" data if appropriate
	- calculates the vectrex scale and strength
	- respects "OPTIMIZED" flag!

	Leaves non zeroed!
*/
void v_directDraw32(int32_t xStart, int32_t yStart, int32_t xEnd, int32_t yEnd, uint8_t brightness)
{
	xStart = xStart*sizeX+offsetX;
	yStart = yStart*sizeY+offsetY;
	xEnd = xEnd*sizeX+offsetX;
	yEnd = yEnd*sizeY+offsetY;

	v_setBrightness(brightness);
	UNZERO(); // ensure vector beam can be moves

	// this will be the delta values for drawing
	int32_t xDrawDif;
	int32_t yDrawDif;

	// how far away is the cursor from the position we want to start drawing?
	int32_t xMoveDif = xStart-currentCursorX;
	int32_t yMoveDif = yStart-currentCursorY;

	// if not optimized, we always reposition!
	// test if the position of the last end - and the current start differs by more than our set margin
	if (((ABS(xMoveDif) > POSITION_MARGIN) || (ABS(yMoveDif) > POSITION_MARGIN) ) || (optimizationON==0))
	{
		// determine whether to zero and reposition (rather than do a move from here)
		// three facters
		// - did we already have to many consecutive moves? If so drift will be large - and it is better to zero once in a while
		// - will the next scale be considerably different from the current one? - if so - better start anew
		// - is the new position very far away from the old, if so, it might even be faster to zero and reposition
		//   (possibly redundant to reason two)

		int resetPos = consecutiveDraws > MAX_CONSECUTIVE_DRAWS;
		resetPos += ABS((currentScale-GET_OPTIMAL_SCALE(xMoveDif, yMoveDif)) > 20);
		resetPos += ((ABS(xMoveDif)>7000) || (ABS(yMoveDif)>7000) );

		// if not optimized, we always reposition!
		resetPos += optimizationON==0;
		if (resetPos)
		{
			// reset to zero - and position absolut
			ZERO_AND_WAIT();
			UNZERO();

			// reposition beam
			v_directDeltaMove32start(xStart, yStart);
			consecutiveDraws = 0;
			xDrawDif = xEnd-xStart;
			yDrawDif = yEnd-yStart;
			SET_OPTIMAL_SCALE(xDrawDif, yDrawDif);
			WAIT_T1_END_LAST();//v_directDeltaMoveEnd();
		}
		else
		{
			// position relative
			// reposition beam
			v_directDeltaMove32start(xMoveDif, yMoveDif);
			consecutiveDraws++;
			xDrawDif = xEnd-xStart;
			yDrawDif = yEnd-yStart;
			SET_OPTIMAL_SCALE(xDrawDif, yDrawDif);
			WAIT_T1_END_LAST();//v_directDeltaMoveEnd();
		}
	}
	else
	{
		// no repositioning -> draw directly again!
		xDrawDif = xEnd-xStart;
		yDrawDif = yEnd-yStart;
		SET_OPTIMAL_SCALE(xDrawDif, yDrawDif);
	}
	SET_YSH16(yDrawDif);
	SET_XSH16(xDrawDif);
	START_T1_TIMER();
  SWITCH_BEAM_ON();
	consecutiveDraws++;
	currentCursorX = xEnd;
	currentCursorY = yEnd;

	WAIT_T1_END();
	SWITCH_BEAM_OFF();
}

/***********************************************************************/

/*	Draws "directly" (not buffered) to vectrex screen.
	Draws from the current position to the given delta values.
 	Assume delta values are in the range: -32768 - +32767,
	assuming 0,0 is in the middle of the screen.
 	"32" name tag denotes signed long range.

	This function acts "smart", it investigates given data (in regards to
    	last used data) and determines:
	- calculates the vectrex scale and strength

	Leaves non zeroed!
*/
void v_directDeltaDraw32(int32_t _xLen, int32_t _yLen, uint8_t brightness)
{
	int32_t xLen = _xLen*sizeX;
	int32_t yLen = _yLen*sizeY;

  
  v_setBrightness(brightness);
	SET_OPTIMAL_SCALE(xLen, yLen);

	SET_YSH16(yLen);
	SET_XSH16(xLen);
	SWITCH_BEAM_ON();
	START_T1_TIMER();
	currentCursorX += _xLen;
	currentCursorY += _yLen;

	// if we have enough space and a correct pi timer
	// than we could wait with a PI timer in parallel and
	// resume "earlier", and switch the beam off
	// at EXACTLY the moment the timer finishes
	// lookup wait table 0x00 - 0xff (scale) for 100% correct timer values?
	WAIT_T1_END();
	DELAY_CYCLES(1); // TEST: Gyrocks needs one additional cycle - seemingly some other VIA register is other set to soon. (Hershey Fonts))
	SWITCH_BEAM_OFF();
}

/***********************************************************************/

/* This routine resets the integrator offsets to zero. This might be neccessary once in a while.
 *
 * The "opposite" (apart from natural degradation of the offsets) is Kristof Tuts like calibration.
 * Which actively sets a integrator offset to a very small value to compensate vectrex "drift".
 */
void v_resetIntegratorOffsets()
{
	SET (VIA_port_b, 0x81);
	DELAY_PORT_B_BEFORE_PORT_A();
	SET (VIA_port_a, 0x00);
	DELAY_CYCLES(4);
	SET (VIA_port_b, 0x80);
	DELAY_CYCLES(6);
	if (calibrationValue==0)
	{
		// reset integrators
		SET (VIA_port_b, 2);	// mux=1, enable mux - integrator offset = 0
		DELAY_CYCLES(6);
		SET (VIA_port_b, 1);	// disable mux
	}
	else
	{
		SET (VIA_port_b, 0x82);
		DELAY_PORT_B_BEFORE_PORT_A();
		SET (VIA_port_a, calibrationValue);
		DELAY_CYCLES(6);
		SET (VIA_port_a, 0xff);
		DELAY_CYCLES(2);
		SET (VIA_port_b, 0x83);
	}
	DELAY_CYCLES(4);
	currentPortA=0x100;// non regular value!
}

/***********************************************************************/

/* This routine moves the vector beam to the extrem edges of the screen and zeros afterwards
 * this prevents "vector collaps".
 * If the "application" uses vectors which are far from the center anyway... this is not needed.
 */
void v_deflok()
{
	ZERO_AND_WAIT();
	UNZERO();
	v_setScale(255);
	SET_YSH_IMMEDIATE_8(127);
	SET_XSH_IMMEDIATE_8(127);
	START_WAIT_T1();
	ZERO_AND_WAIT();
	UNZERO();
	SET_YSH_IMMEDIATE_8(-127);
	SET_XSH_IMMEDIATE_8(-127);
	START_WAIT_T1();
	ZERO_AND_CONTINUE();
}

/***********************************************************************/

/* 	Set vectrex refresh rate for T2 "WaitRecal".
	Parameter in in Hz.
	-> so 50 Hz -> results in  Vec_Rfrsh = 0x3075
*/
void v_setRefresh(int hz)
{
	int32_t cycles = 1500000 / hz; // Hz to vectrex cycles
	Vec_Rfrsh = ((cycles&0xff)*256) + ((cycles>>8)&0xff); // swap hi / lo bytes for little endian VIA
}

/***********************************************************************/

/* 	A vectrex BIOS like function that can be called "once per round".
   	This function waits till the Vectrex Timer T2 is finished (timing can be set
   	using "v_setRefresh(int hz)".

	After the wait finishes, following things are further executed:
	- the Timer T2 will be reset to the set frequency.
	- v_deflok()
	- v_resetIntegratorOffsets()

	Leaves with scale set to 0xff.
	
	Resets ioDone (so that joystick reads are only done once per round).
	
	Also checks and reacts on PiTrex "button" sets (return to boot screen, calibrate etc)
 */
void v_WaitRecal()
{
	ioDone = 0;
	SWITCH_BEAM_OFF();
#ifdef PITREX_DEBUG
	if (currentButtonState ==0x9) // button 1 + 4 pressed - print speed info
	{
		uint16_t timeLeft = ((Vec_Rfrsh&0xff)*256+(Vec_Rfrsh>>8)) - GET(VIA_t2_cnt_hi)*256;
		printf("Refresh usage: %i%% \r\n", (timeLeft*100/((Vec_Rfrsh&0xff)*256+(Vec_Rfrsh>>8)))  );
	}
#endif

	v_resetDetection();
	if (currentButtonState ==0xf) // button 1+ 2 + 3+4 -> go menu
	{
#ifndef FREESTANDING
	exit(0); // pressing all four buttons exits
#endif
	      printf("Restarting kernel...\r\n");
	      uint32_t progSapce = 0xb00000;//0x1f000000;
	      void (*progStart)(void) = (void (*)(void))progSapce;
	      progStart();
	}

	if (currentButtonState ==0x6) // button 2 + 3 pressed - enter calibration menu1
	  inCalibration = 1;
	if (inCalibration)
	 	v_calibrate();

	// wait for Via T2 to expire -> 50Hz
	while ((GET (VIA_int_flags) & 0x20) == 0)
	{
		; // might do a usleep(1); here
	}

	// reset T2 VIA timer to 50Hz
	SETW (VIA_t2, Vec_Rfrsh);
	currentZSH = 0x100;
	consecutiveDraws = 0;
	v_deflok();
	v_resetIntegratorOffsets();
}

/***********************************************************************/
// detect VIA reset
// TODO
//-> this might work, if we don't use T1, but our own PI timer
// which atm we do...
void v_resetDetection()
{
	if ((GET(VIA_DDR_a) == 0) && (GET(VIA_DDR_b) == 0))
	{
	  printf("Reset detected!\r\n");
	      // wait till reset is finished
	      while ((GET(VIA_DDR_a) == 0) && (GET(VIA_DDR_b) == 0))
	      {
			SET(VIA_DDR_a, 0xff);
			DELAY_CYCLES(4);
	      }
	      printf("Restarting kernel...\r\n");
	      uint32_t progSapce = 0xb00000;//0x1f000000;
	      void (*progStart)(void) = (void (*)(void))progSapce;
	      progStart();
	}
}

	
/***********************************************************************/

/* 	Reads the vectrex button states (port 1 and port 2).
 	Needs only be called once per round (or even only every second round)
	result also in variable: currentButtonState
		bit 0 = button 1 port 1
		bit 1 = button 2 port 1
		bit 2 = button 3 port 1
		bit 3 = button 4 port 1
		bit 4 = button 1 port 2
		bit 5 = button 2 port 2
		bit 6 = button 3 port 2
		bit 7 = button 4 port 2
 	if bit == 1, than button is pressed
 	if bit == 0, than button is not pressed
*/
uint8_t v_readButtons()
{
  if (ioDone & V_BUTTONS_READ) return currentButtonState;
  ioDone |= V_BUTTONS_READ;
	// read of buttons goes thru the PSG sound chip, PSG port A
	//
	SET(VIA_port_a, 0x0e); // prepare access of psg port A (0x0e) by writing the register value to VIA port A
	DELAY_CYCLES(4);
	SET(VIA_port_b, 0x99); // set VIA port B to settings: sound BDIR on, BC1 on, mux off
	DELAY_CYCLES(4);
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)
	DELAY_CYCLES(4);

	SET(VIA_DDR_a, 0x00); // set VIA DDR A to input
	DELAY_CYCLES(4);
	SET(VIA_port_b, 0x89); // set VIA port B to settings: sound BDIR on, BC1 on, mux off
	DELAY_CYCLES(6);
	currentButtonState = ~GET(VIA_port_a); // Read buttons
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)
	DELAY_CYCLES(4);
	SET(VIA_DDR_a, 0xff); // set VIA DDR A to output
	DELAY_CYCLES(4);
	currentPortA = 0x100;
	return currentButtonState;
}

/***********************************************************************/

/* 	Reads the vectrex joystick states (port 1) DIGITAL.
	(Faster than analog call!)
 	Needs only be called once per round (or even only every second round)
  	result in variable:
		currentJoy1Y = -1 down
		currentJoy1Y = 0 centered
		currentJoy1Y = 1 up

		currentJoy1X = -1 left
		currentJoy1X = 0 centered
		currentJoy1X = 1 right
*/

void v_readJoystick1Digital()
{
  if (ioDone & V_JOY_DIGITAL_READ) return;
  ioDone |= V_JOY_DIGITAL_READ;

	// Y ANALOG
	SET(VIA_port_a, 0x00); // clear VIA port A
	DELAY_CYCLES(4);
	SET(VIA_port_b, 0x02); // set VIA port B mux enabled, mux sel = 01 (vertical pot port 0)

	// wait for joystick comparators to "settle"
	DELAY_CYCLES(60); // must be tested! can probably be less?

	currentJoy1Y = -1; // default down
	SET(VIA_port_b, 0x03); // set VIA port B mux
	DELAY_PORT_B_BEFORE_PORT_A();
	SET(VIA_port_a, 0x40); // load a with test value (positive y), test value to DAC
	DELAY_CYCLES(4);
	if ((GET(VIA_port_b) & 0x20) == 0x20)
	{
		currentJoy1Y = 1; //up
	}
	else
	{
		SET(VIA_port_a, -0x40); // load a with test value (negative y), test value to DAC
		DELAY_CYCLES(4);
		if ((GET(VIA_port_b) & 0x20) == 0x20)
		{
			currentJoy1Y = 0; // no direction
		}
	}

	// X ANALOG
	SET(VIA_port_a, 0x00); // clear VIA port A
	DELAY_CYCLES(4);
	SET(VIA_port_b, 0x00); // set VIA port B mux enabled, mux sel = 00 (horizontal pot port 0)
	// wait for joystick comparators to "settle"
	DELAY_CYCLES(60); // must be tested! can probably be less?

	currentJoy1X = -1; // default left
	SET(VIA_port_b, 0x03); // set VIA port B mux
	DELAY_PORT_B_BEFORE_PORT_A();
	SET(VIA_port_a, 0x40); // load a with test value (positive y), test value to DAC
	DELAY_CYCLES(2);
	if ((GET(VIA_port_b) & 0x20) == 0x20)
	{
		currentJoy1X = 1; //right
	}
	else
	{
		DELAY_CYCLES(4);
		SET(VIA_port_a, -0x40); // load a with test value (negative y), test value to DAC
		DELAY_CYCLES(4);
		if ((GET(VIA_port_b) & 0x20) == 0x20)
		{
			currentJoy1X = 0; // no direction
		}
		DELAY_CYCLES(4);
	}
	// set port A reference value to unkown
	currentYSH = currentPortA=0x100; // reset saved current values to unkown state
	v_resetIntegratorOffsets();
}

/***********************************************************************/

/* 	Reads the vectrex joystick states (port 1) ANALOG.
	(Slower than digital call!)
 	Needs only be called once per round (or even only every second round)
  	result in variable:
		currentJoy1Y = -1 - -127 down
		currentJoy1Y = 0 centered
		currentJoy1Y = +1 - +127 up

		currentJoy1X = -1 - -127  left
		currentJoy1X = 0 centered
		currentJoy1X = +1 - +127 right
*/
void v_readJoystick1Analog()
{
  if (ioDone & V_JOY_ANALOG_READ) return;
  ioDone |= V_JOY_ANALOG_READ;

	SET(VIA_port_a, 0x00); // clear VIA port A
	DELAY_CYCLES(4);
	SET(VIA_port_b, 0x02); // set VIA port B mux enabled, mux sel = 01 (vertical pot port 0)
	// wait for joystick comparators to "settle"
	DELAY_CYCLES(60); // must be tested! can probably be less?

	currentJoy1Y = 0; // default centered
	SET(VIA_port_b, 0x03); // set VIA port B mux disabled
	DELAY_PORT_B_BEFORE_PORT_A();
	uint8_t compareBit = 0x80;
	currentJoy1Y = 0;

	do
	{
		SET(VIA_port_a, currentJoy1Y); // load a with test value (positive y), test value to DAC
		DELAY_CYCLES(4);
		if ((GET(VIA_port_b) & 0x20) == 0)
		{
			currentJoy1Y = currentJoy1Y ^ compareBit;
		}
		DELAY_CYCLES(4);
		compareBit = compareBit>>1;
		currentJoy1Y = currentJoy1Y | compareBit;
	} while (compareBit!=0);

	SET(VIA_port_a, 0x00); // clear VIA port A
	DELAY_CYCLES(4);
	SET(VIA_port_b, 0x00); // set VIA port B mux enabled, mux sel = 01 (horizontal pot port 0)
	// wait for joystick comparators to "settle"
	DELAY_CYCLES(60); // must be tested! can probably be less?

	currentJoy1X = 0; // default centered
	SET(VIA_port_b, 0x01); // set VIA port B mux disabled
	DELAY_PORT_B_BEFORE_PORT_A();
	compareBit = 0x80;
	currentJoy1X = 0;

	do
	{
		SET(VIA_port_a, currentJoy1X); // load a with test value (positive y), test value to DAC
		DELAY_CYCLES(4);
		if ((GET(VIA_port_b) & 0x20) == 0)
		{
			currentJoy1X = currentJoy1X ^ compareBit;
		}
		DELAY_CYCLES(4);
		compareBit = compareBit>>1;
		currentJoy1X = currentJoy1X | compareBit;
	} while (compareBit!=0);

	// set port A reference value to unkown
	currentYSH = currentPortA=0x100; // reset saved current values to unkown state
	v_resetIntegratorOffsets();
}

/***********************************************************************/

void v_writePSG_buffered(uint8_t reg, uint8_t data)
{
	if (psgShadow[reg] == data) return;
	psgShadow[reg] = data;

	SET(VIA_port_a, reg); // prepare access of psg port A (0x0e) by writing the register value to VIA port A
	SET(VIA_port_b, 0x99); // set VIA port B to settings: sound BDIR on, BC1 on, mux off
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)

	SET(VIA_port_a, data); // write data to port a of via -> and than to psg
	SET(VIA_port_b, 0x91); // set VIA port B to settings: sound BDIR on, BC1 on, mux off, write to PSG
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)
	currentPortA = 0x100; // undefined
}
void v_writePSG(uint8_t reg, uint8_t data)
{
	SET(VIA_port_a, reg); // prepare access of psg port A (0x0e) by writing the register value to VIA port A
	SET(VIA_port_b, 0x99); // set VIA port B to settings: sound BDIR on, BC1 on, mux off
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)

	SET(VIA_port_a, data); // write data to port a of via -> and than to psg
	SET(VIA_port_b, 0x91); // set VIA port B to settings: sound BDIR on, BC1 on, mux off, write to PSG
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)
	currentPortA = 0x100; // undefined
}

/***********************************************************************/

uint8_t v_readPSG(uint8_t reg)
{
	SET(VIA_port_a, reg); // prepare access of psg port A (0x0e) by writing the register value to VIA port A
	SET(VIA_port_b, 0x99); // set VIA port B to settings: sound BDIR on, BC1 on, mux off
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)

	SET(VIA_DDR_a, 0x00); // set VIA DDR A to input
	SET(VIA_port_b, 0x89); // set VIA port B to settings: sound BDIR on, BC1 on, mux off, read from psg
	uint8_t directData = GET(VIA_port_a); // Read buttons
	SET(VIA_port_b, 0x81); // set VIA Port B = 81, mux disabled, RAMP disabled, BC1/BDIR = 00 (PSG inactive)
	SET(VIA_DDR_a, 0xff); // set VIA DDR A to output
	currentPortA = 0x100; // undefined
  return directData;
}

/***********************************************************************/

/* 	Sets the Vectrex Zero state - and waits for zeroing to have full effect.
	Zeroing is NOT disabled upon end of function!
*/
void v_zeroWait()
{
	ZERO_AND_WAIT();
}

/***********************************************************************/

/* 	Sets the current scale value absolut to given value.
    Does NOT change VIA!
*/
void v_setScale(uint16_t s)
{
  lastScale = currentScale;
  currentScale = s;
}

/***********************************************************************/

/* 	Moves from current location to
	delta given by parameters.
	- does not set any internal variables
	- does not change scale
	- waits till position is reached

	Caller must ensure zeroing and similar states are taken care of beforehand!
*/
void v_moveToImmediate8(int8_t xLen, int8_t yLen)
{
	SET_YSH_IMMEDIATE_8(yLen);
	SET_XSH_IMMEDIATE_8(xLen);
	START_T1_TIMER();
	WAIT_T1_END();
}

/***********************************************************************/

/* 	Draws from current location to
	delta given by parameters.
	- does not set any internal variables
	- does not change scale
	- waits till position is reached

	Caller must ensure zeroing and similar states are taken care of!
*/
void v_drawToImmediate8(int8_t xLen, int8_t yLen)
{
	SET_YSH_IMMEDIATE_8(yLen);
	SET_XSH_IMMEDIATE_8(xLen);
	SWITCH_BEAM_ON();
	START_T1_TIMER();
	WAIT_T1_END();
	SWITCH_BEAM_OFF();
}

/***********************************************************************/
/* All things calibration follows                                      */
/***********************************************************************/
// not used anymore - but can still be "looked at"
void v_delayCalibration()
{
	if ((currentButtonState&0x08) == (0x08)) calibrationStep=0;
	int testScale = 0x70;

	v_setBrightness(0x5f);
	v_zeroWait();
	v_setScale(testScale);
	UNZERO();
	v_moveToImmediate8(-64, 0);
	v_drawToImmediate8(127, 0);

	v_deflok();
	v_resetIntegratorOffsets();
	v_zeroWait();
	v_setScale(testScale);
	UNZERO();
	v_moveToImmediate8(-64, -10);

	v_setScale(0xff);

	SET_YSH_IMMEDIATE_8(0);
	SET_XSH_IMMEDIATE_8(127);
	SWITCH_BEAM_ON();
	START_T1_TIMER();
	v_delayCycles(testScale);
	DELAY_CYCLES(DELAY_AFTER_T1_END_VALUE); // to have comparable values we must "wait" the same time as for T1
	SWITCH_BEAM_OFF();
	WAIT_T1_END();
}

/***********************************************************************/

void v_driftCalibration()
{
	if ((currentButtonState&0x08) == (0x08)) calibrationStep=0;
	if (currentJoy1X>80)
	{
		calibrationValue++;
	}
	if (currentJoy1X<-80)
	{
		calibrationValue--;
	}

	char *message = "JOYSTICK LEFT <> RIGHT!";
	v_printString(-50, 20, message, 5);

#ifdef PITREX_DEBUG
	printf("drift: %i (%i)\r\n", calibrationValue, currentJoy1X);
#endif
}

/***********************************************************************/
// left is smaller X
// down is smaller Y
void v_sizeCalibration()
{
	if ((currentButtonState&0x08) == (0x08)) calibrationStep=0;
	if (currentJoy1X>80)
	{
		sizeX+= 0.01;
	}
	if (currentJoy1X<-80)
	{
		sizeX-= 0.01;
	}
	if (currentJoy1Y>80)
	{
		sizeY+= 0.01;
	}
	if (currentJoy1Y<-80)
	{
		sizeY-= 0.01;
	}

	char *message = "JOYSTICK";

	v_printString(-50, 20, message, 5);
#ifdef PITREX_DEBUG
	printf("stretch x, y: %f, %f\r\n", sizeX, sizeY);
#endif

}

/***********************************************************************/

void v_positionCalibration()
{
	if ((currentButtonState&0x08) == (0x08)) calibrationStep=0;
	if (currentJoy1X>80)
	{
		offsetX+= 100;
	}
	if (currentJoy1X<-80)
	{
		offsetX-= 100;
	}
	if (currentJoy1Y>80)
	{
		offsetY+= 100;
	}
	if (currentJoy1Y<-80)
	{
		offsetY-= 100;
	}

	char *message = "JOYSTICK";
	v_printString(-50, 20, message, 5);

#ifdef PITREX_DEBUG
	printf("offset x, y: %i, %i\r\n", offsetX, offsetY);
#endif
}

/***********************************************************************/

void v_calibrationMenu()
{
	char *menu1 = "DELAY";
	char *menu2 = "DRIFT";
	char *menu3 = "SIZE";
	char *menu4 = "POSITION";
	char *menu5on = "OPTIMZE ON";
	char *menu5off = "OPTIMZE OFF";
	char *menu6 = "DONE";
	char *selected = ">";

	v_printString(-50, 70, menu1, 5);
	v_printString(-50, 60, menu2, 5);
	v_printString(-50, 50, menu3, 5);
	v_printString(-50, 40, menu4, 5);
	if (optimizationON)
		v_printString(-50, 30, menu5on, 5);
	else
		v_printString(-50, 30, menu5off, 5);
	v_printString(-50, 20, menu6, 5);

	v_printString(-60, 20+10*selectedMenu, selected, 5);

	if ((currentJoy1Y>50) && (selectionMade==0))
	{
		selectedMenu++;
		if (selectedMenu>5) selectedMenu=5;
		selectionMade = 1;
	}
	if ((currentJoy1Y<-50) && (selectionMade==0))
	{
		selectedMenu--;
		if (selectedMenu<0) selectedMenu=0;
		selectionMade = 1;
	}
	if (ABS(currentJoy1Y)<20) selectionMade =0;
	if ((currentButtonState&0x04) == (0x04))
	{
		if (selectedMenu==0)
			inCalibration=0;
		else if (selectedMenu==1)
			optimizationON = optimizationON?0:1;
		else
			calibrationStep = selectedMenu;
	}
}

/***********************************************************************/

/* 	Calibration "switcher", called from "WaitRecal"
*/
void v_calibrate()
{
	char *message = "CALIBRATION";
	v_printString(-50, 100, message, 10);
	v_readButtons();
	v_readJoystick1Analog();
	switch (calibrationStep)
	{
		case 0:
		{
			v_calibrationMenu();
			break;
		}
		case 2:
		{
			v_positionCalibration();
			break;
		}
		case 3:
		{
			v_sizeCalibration();
			break;
		}
		case 4:
		{
			v_driftCalibration();
			break;
		}
		case 5:
		{
			v_delayCalibration();
			break;
		}
		default:
		{
			inCalibration=0;
		}
	}
}

/***********************************************************************/
#ifdef FREESTANDING
#include <baremetal/rpi-systimer.h>
static rpi_sys_timer_t* rpiSystemTimer = (rpi_sys_timer_t*)RPI_SYSTIMER_BASE;
uint32_t markStartLo = 0;
uint32_t markStartHi = 0;
uint32_t markRollOverExpected = 0;



/* sets a "mark" for the system timer
 * returns just after a timer "tick" occured
 */
void setMarkStart(void)
{
    volatile uint32_t ts = rpiSystemTimer->counter_lo;
    // ensure the timer "just" switched
    // the timer has only a resolution of 1 micro second
    // one instruction takes 1 nano second - soooo actually
    // the resolution is only exact for 1000 pi instructions!!!
    markStartHi = rpiSystemTimer->counter_hi;
    while (ts == rpiSystemTimer->counter_lo);
    
    if (markStartHi != rpiSystemTimer->counter_hi) 
      markStartHi = rpiSystemTimer->counter_hi;
    markStartLo = rpiSystemTimer->counter_lo;
    return ;
}

/* waits till # of microseconds passed since mark was set */
void waitMarkEnd(uint32_t offsetMicro)
{
    volatile uint32_t ts = markStartLo + offsetMicro;
//  printf("mMarkEnd: %i, %i, %i\r\n", ts, rpiSystemTimer->counter_lo, offsetMicro);
    if (ts < markStartLo)
    {
        // wait till rollover
        while (markStartHi == rpiSystemTimer->counter_hi)
	  ;
    }
        
    // return when the timer has "just" switched
    while (ts > rpiSystemTimer->counter_lo) ;
    return;
}
/* waits till the next "full" micro second tick is done
   then returns the full tick counter since last mark */
int waitFullMicro()
{
    int ticks =  rpiSystemTimer->counter_lo-markStartLo+1;
    
    volatile uint32_t ts = rpiSystemTimer->counter_lo;
        
    // return when the timer has "just" switched
    while (ts == rpiSystemTimer->counter_lo) ;
    return ticks;
}

/* waits till # of microseconds passed since mark was set */
/* and sets new mark */
void waitMarkEndMark(uint32_t offsetMicro)
{
    volatile uint32_t ts = markStartLo + offsetMicro;
    if (ts < markStartLo)
    {
        // wait till rollover
        while (markStartHi == rpiSystemTimer->counter_hi) ;
    }
        
    // return when the timer has "just" switched
    while (ts > rpiSystemTimer->counter_lo) ;
    markStartHi = rpiSystemTimer->counter_hi;
    markStartLo = rpiSystemTimer->counter_lo;
    return;
}

// obsolete!
// -> the same as "waitMarkEnd" now
//
/* Waits till # of vectrex cycles have passed
 * uses Systimer for "full" microseconds
 * and Nano Wait for partials.
 * 
 * One Vectrex cycle = 0,666 useconds
 * 
 * ...passed since mark was set */
void waitCycleMarkEnd(uint32_t cycles)
{
 //   markStartLo = 0;
//  if(markStartLo==0) return;
/*
  uint32_t micros = cycles * 0.840;
    uint32_t nanos = cycles * 840 - micros*1000;
    waitMarkEnd(micros);
    WAIT_CYCLE_NANO(nanos);
*/   
    waitMarkEnd(cycles);
    return;
}

void measureTime()
{
  SET(VIA_t1_cnt_lo, 100); 
  DELAY_CYCLES(4);
  setMarkStart();
  SET(VIA_t1_cnt_hi, 0); 
  while ((GET (VIA_int_flags) & 0x40) == 0);
  
  uint32_t lmarkStartHi = rpiSystemTimer->counter_hi;
  if (lmarkStartHi != rpiSystemTimer->counter_hi) 
    lmarkStartHi = rpiSystemTimer->counter_hi;
  uint32_t lmarkStartLo = rpiSystemTimer->counter_lo;
//  printf("LoDif: %i\r\n", lmarkStartLo-markStartLo);
}

// only using low counters!
uint32_t v_millis()
{
    uint32_t micros= rpiSystemTimer->counter_lo;
	return micros/1000;
}

// only using low counters!
uint32_t v_micros()
{
	return rpiSystemTimer->counter_lo;
}

#endif

void v_initYM(unsigned char *b, uint16_t length, int l)
{
	ymBuffer = b;
	ymloop = l;
	ymLength = length;
	ymPos = 0;
}

int v_playYM()
{
	if (ymPos>=ymLength) 
	{
	  if (ymloop==1)
	    ymPos = 0;
	    else
	    return 0;
	}
	unsigned char *currentPointer = ymBuffer+ymPos*14;
	ymPos++;
	for (int i=0; i<14; i++)
	{
		v_writePSG_buffered(i, *currentPointer);
		currentPointer++;
	}
	return 1;
}

// does play a complete sample on vectrex, does not return in between... nothing is displayed!
// sample rate is e.g. 15000
// one sample each 0,000066 seconds
// = 0,0666 milli seconds
// = each 66 micro seconds
void v_playDirectSampleAll(char *ymBufferLoad, int fsize, int rate)
{
  int counter = 0;
  int microsToWait = (1000*1000) / rate;
  
  SET(VIA_port_a, 0); 
  DELAY_CYCLES(4);
   
  SET(VIA_port_b, 0x06); 
  DELAY_CYCLES(4);
  while (counter < fsize)
  {
    char b = *ymBufferLoad++;
    SET(VIA_port_a, b); 
    int endTime = v_micros()+microsToWait;
    counter++;
    while (v_micros()<endTime) ;
  }
    SET(VIA_port_b, 0x01); 
    DELAY_CYCLES(4);
}

// blocking error message!
void v_error(char *message)
{
  printf("Vectrex/PiTrex bailing out - error:\r\n\t%s\r\n" , message);
  while (1)
  {
    v_WaitRecal();
    v_setBrightness(64);        /* set intensity of vector beam... */
    v_printStringRaster(-30, 80, "ERROR", 40, -7, 0);
    v_printStringRaster(-100, 0, message, 40, -7, 0);
    v_readButtons();
  }
}

