#ifndef _VECTREX_INTERFACE_H
#define _VECTREX_INTERFACE_H 1
#include <stddef.h>
#include <stdint.h>

void v_init();
void v_delayCycles(uint32_t cycles);
void v_delayCyclesQuarter();
void v_delayCyclesEighth();
void delayNano(uint32_t n);
void setMarkStart(void);
void waitMarkEnd(uint32_t offsetMicro);
void waitMarkEndMark(uint32_t offsetMicro);
void waitCycleMarkEnd(uint32_t cycles);
void measureTime(); // debug and measure only

void v_setBrightness(uint8_t brightness);
void v_directDeltaMove32start(int32_t xLen, int32_t yLen);
void v_directDeltaMoveEnd();
void v_directDraw32(int32_t xStart, int32_t yStart, int32_t xEnd, int32_t yEnd, uint8_t brightness);
void v_directDeltaDraw32(int32_t xLen, int32_t yLen, uint8_t brightness);
void v_setRefresh(int hz);
void v_zeroWait();
void v_deflok();
void v_resetIntegratorOffsets();
void v_setRefresh(int hz);
void v_WaitRecal();
void resetDetection();
void v_calibrate();
uint8_t v_readButtons();
void v_readJoystick1Digital();
void v_readJoystick1Analog();
void v_printString(int8_t x, int8_t y, char* string, uint8_t textSize);

void v_setScale(uint16_t s);
void v_moveToImmediate8(int8_t xLen, int8_t yLen);
void v_drawToImmediate8(int8_t xLen, int8_t yLen);
void v_writePSG(uint8_t reg, uint8_t data);
uint8_t v_readPSG(uint8_t reg);


int32_t currentCursorX; // 32 bit positioning value X, "normal" values are 16 bit,
						// but enlarging the screen may leed to higher values
int32_t currentCursorY; // 32 bit positioning value Y
uint16_t currentScale; // currently active scale factor (T1 timer in VIA)
uint16_t lastScale; // last active scale factor (T1 timer in VIA)
int16_t consecutiveDraws; // how many lines/moves were made after another!
uint8_t currentButtonState;
int8_t currentJoy1X;
int8_t currentJoy1Y;
int8_t currentJoy2X;
int8_t currentJoy2Y;
int16_t currentPortA; // == portA (also X SH)
int16_t currentYSH; // Y SH
int16_t currentZSH; // Brightness

int8_t isCranky; // cranky should be checked during calibration!

unsigned int Vec_Rfrsh;//0x3075;//12405; // 30000 = $7530 little endian = 3075 = 12405 50U;	// PLACEHOLDER FOR NOW -- FIX!
int Vec_Loop_Count;

#include <pitrex/via6522.h>

#ifndef VIA_port_b
#define VIA_port_b    ((unsigned short) 0xD000)	// VIA port B data I/O register
#endif
#ifndef VIA_port_a
#define VIA_port_a    ((unsigned short) 0xD001)	// VIA port A data I/O register (handshaking)
#endif
#ifndef VIA_DDR_b
#define VIA_DDR_b     ((unsigned short) 0xD002)	// VIA port B data direction register (0=input 1=output)
#endif
#ifndef VIA_DDR_a
#define VIA_DDR_a     ((unsigned short) 0xD003)	// VIA port A data direction register (0=input 1=output)
#endif
#ifndef VIA_t1
#define VIA_t1        ((unsigned short) 0xD004)	// VIA timer 1 count
#endif
#ifndef VIA_t1_cnt_lo
#define VIA_t1_cnt_lo ((unsigned short) 0xD004)	// VIA timer 1 count register lo (scale
#endif
#ifndef VIA_t1_cnt_hi
#define VIA_t1_cnt_hi ((unsigned short) 0xD005)	// VIA timer 1 count register hi
#endif
#ifndef VIA_t2
#define VIA_t2        ((unsigned short) 0xD008)	// VIA timer 2 count register BOTH BYTES
#endif
#ifndef VIA_t2_cnt_lo
#define VIA_t2_cnt_lo ((unsigned short) 0xD008)	// VIA timer 2 count register lo
#endif
#ifndef VIA_t2_cnt_hi
#define VIA_t2_cnt_hi ((unsigned short) 0xD009)	// VIA timer 2 count register hi
#endif
#ifndef VIA_shift_reg
#define VIA_shift_reg ((unsigned short) 0xD00A)	// VIA shift register
#endif
#ifndef VIA_aux_cntl
#define VIA_aux_cntl  ((unsigned short) 0xD00B)	// VIA auxiliary control
#endif
#ifndef VIA_cntl
#define VIA_cntl      ((unsigned short) 0xD00C)	// VIA control register
#endif
#ifndef VIA_int_flags
#define VIA_int_flags ((unsigned short) 0xD00D)	// VIA interrupt flags register
#endif
#ifndef VIA_int_enable
#define VIA_int_enable ((unsigned short) 0xD00E)	// VIA interrupt enable register
#endif

// a version of these that does not evaluate the arguments twice would be preferred...
#ifndef MAX
#define MAX(A,B) ((A)>(B)?(A):(B))
#endif
#ifndef ABS
#define ABS(A) ((A)>0?(A):(-(A)))
#endif

#define SETW_inverse(address, val) do {vectrexwrite((address), (val)&255);vectrexwrite((address)+1, ((val)>>8)&255); } while(0)
#define SETW(address, val) do {vectrexwrite(address, ((val)>>8)&255); vectrexwrite((address)+1, (val)&255);} while(0)
#define SET(address, val) vectrexwrite(address, val)
#define GET(address) vectrexread(address)

#define V_BUTTONS_READ 1
#define V_JOY_DIGITAL_READ 2
#define V_JOY_ANALOG_READ 4

#define MAX_USED_STRENGTH 127//27 // abs
#define SCALE_STRENGTH_DIF 2//4 //
#define MINSCALE 5 //15 to low might interfere with switching shift states to soon -> shift stalling
// #define MAXSCALE 255 not used, AND can be larger than 255
#define MAX_CONSECUTIVE_DRAWS 65






#define POSITION_MARGIN 10 // margin from where two positions are taken as equal 16bit

// delay value in vectrex cycles
#ifdef FREESTANDING
#define DELAY_NANO(n) delayNano((n))
#define DELAY_ZERO_VALUE 70 // probably less, this can be adjusted, by max x position, the nearer to the center the less waits
#define DELAY_AFTER_T1_END_VALUE 15//12
#define DELAY_AFTER_XSH_VALUE 0//0 // probably 0
#define DELAY_AFTER_YSH_VALUE 8//6 might work too
#define DELAY_CRANKY_VALUE 3 //- less, when less max strength is used!
#define DELAY_PORT_B_BEFORE_PORT_A_VALUE 2 // not sure is this is a thing
#ifdef MHZ1000
#define DELAY_PI_CYCLE_EQUIVALENT 940//940 // nano seconds
#else
#define DELAY_PI_CYCLE_EQUIVALENT 660 // nano seconds
#endif
#else
#define DELAY_ZERO_VALUE 70 // probably less, this can be adjusted, by max x position, the nearer to the center the less waits
#define DELAY_AFTER_T1_END_VALUE 12//12
#define DELAY_AFTER_XSH_VALUE 0 // probably 0
#define DELAY_AFTER_YSH_VALUE 8//4 //
#define DELAY_CRANKY_VALUE 3 //- less, when less max strength is used!
#define DELAY_PORT_B_BEFORE_PORT_A_VALUE 2 // not sure is this is a thing
#define DELAY_PI_CYCLE_EQUIVALENT 940//40 //930
#endif

/* possibly add:
  DELAY_AFTER_T1_START_VALUE 2 // (my personal stupid vectrex)
  DELAY_AFTER_T1_SET_VALUE 2 // (wait after scale is set - if started to fast??? dunno, is that a thing?)
*/

#define SET_SHIFT_REG(v) SET(VIA_shift_reg, (v))

//#define BEAM_LIGHT_BY_SHIFT 1 // only shift "beam" implemented for now
#define BEAM_LIGHT_BY_CNTL 
// at the moment the beam (~BLANK) is switched using the SHIFT register
// but using other VIA settings the BEAM can also be switched on/off using the
// CNTL register 

#ifdef BEAM_LIGHT_BY_SHIFT
#define SWITCH_BEAM_ON() SET_SHIFT_REG(0xff)
#define SWITCH_BEAM_OFF() SET_SHIFT_REG(0x00)
#endif

#ifdef BEAM_LIGHT_BY_CNTL
#define SWITCH_BEAM_ON() SET(VIA_cntl, 0xee)
#define SWITCH_BEAM_OFF() SET(VIA_cntl, 0xce)
#endif





#define DELAY_CYCLES_QUARTER() v_delayCyclesQuarter()
#define DELAY_CYCLES_EIGHTH() v_delayCyclesEighth()
#define DELAY_CYCLES(cycles) v_delayCycles((cycles))
#define DELAY_PORT_B_BEFORE_PORT_A() do{DELAY_CYCLES(DELAY_PORT_B_BEFORE_PORT_A_VALUE);} while (0)// not sure is this is a thing
#define DELAY_CRANKY() do{if (isCranky) DELAY_CYCLES(DELAY_CRANKY_VALUE);} while (0) // cranky should be checked during calibration!
#define DELAY_T1_OFF() DELAY_CYCLES(DELAY_AFTER_T1_END_VALUE)
#define DELAY_XSH() DELAY_CYCLES(DELAY_AFTER_XSH_VALUE)
#define DELAY_ZSH() DELAY_YSH()
#define DELAY_YSH()  \
	do{DELAY_CYCLES(DELAY_AFTER_YSH_VALUE); \
	DELAY_CRANKY();} while (0)
	
#ifdef FREESTANDING
#define START_T1_TIMER()  \
do { \
	setMarkStart(); \
	if (lastScale == currentScale) \
	{ \
 		SET(VIA_t1_cnt_hi, currentScale>>8);  \
	} \
	else \
	{ \
 		SETW_inverse(VIA_t1, currentScale);  /* scale due to "enlargement" is 16 bit! */ \
	}  \
} while (0)
#define WAIT_T1_END_LAST() do{waitCycleMarkEnd(lastScale); DELAY_T1_OFF();} while (0) 
#define WAIT_T1_END() do{waitCycleMarkEnd(currentScale); DELAY_T1_OFF();} while (0) 
#define START_WAIT_T1() do{START_T1_TIMER();WAIT_T1_END();} while (0)

#else

#define START_T1_TIMER()			\
do { \
	if (lastScale == currentScale) \
	{ \
 		SET(VIA_t1_cnt_hi, currentScale>>8);  \
	} \
	else \
	{ \
 		SETW_inverse(VIA_t1, currentScale);  /* scale due to "enlargement" is 16 bit! */ \
	}  \
} while (0)
#define WAIT_T1_END() do{while ((GET (VIA_int_flags) & 0x40) == 0); DELAY_T1_OFF();} while (0)
#define WAIT_T1_END_LAST WAIT_T1_END 
// this is missing. Is it the same as above?
#define START_WAIT_T1() do{START_T1_TIMER();WAIT_T1_END();} while (0)
#endif



// chose a good scale for the max size given
// depending on "currentScale" (since it also needs time to switch scales)
// and on size, size should be so that
// it should maxVSizeWord/scale == very near to 127 (8 bit max)
//
// attention! for cranky vectrex large (SH values in Y) values (>100?) need additional waits
// when written to Port A of VIA
#define GET_OPTIMAL_SCALE(a,b) ((MAX(ABS(a), ABS(b))+(MAX_USED_STRENGTH-1))/MAX_USED_STRENGTH)
#define SET_OPTIMAL_SCALE(a,b) \
	do {lastScale = currentScale;currentScale = GET_OPTIMAL_SCALE(a,b); \
	if (currentScale <= MINSCALE) currentScale = MINSCALE;\
	} while (0)

#define UNZERO() SET(VIA_cntl, 0xCE) // disable zeroing, otherwise no positioning can be done
#define ZERO_AND_CONTINUE() do{SET(VIA_cntl, 0xCC);currentCursorX = 0; currentCursorY = 0;} while (0)
#define ZERO_AND_WAIT() do{ \
	ZERO_AND_CONTINUE(); \
	DELAY_CYCLES(DELAY_ZERO_VALUE);} while (0)

#define SET_YSH_IMMEDIATE_8(v) do{ \
	SET(VIA_port_b, 0x00); \
	DELAY_PORT_B_BEFORE_PORT_A(); \
	currentYSH=currentPortA=(uint8_t)(v); \
	SET(VIA_port_a, currentPortA); \
	DELAY_YSH();} while (0)

#define SET_XSH_IMMEDIATE_8(v) do{ \
	SET(VIA_port_b, 0x01); \
	DELAY_PORT_B_BEFORE_PORT_A(); \
	currentPortA=(uint8_t)(v); \
	SET(VIA_port_a, currentPortA); \
	DELAY_XSH();} while (0)

// leaves with mux set to y integrator!
// sets 16 bit values divided by current scale!
// actually: due to possible display scaling, input values can now be larger 16bit
#define SET_YSH16(v) \
  do { \
	int16_t t = (uint8_t)((v)/(currentScale+SCALE_STRENGTH_DIF));  \
	if (t != currentYSH) \
	{ \
		if (t != currentPortA) \
		{ \
			SET(VIA_port_b, 0x00); \
			DELAY_PORT_B_BEFORE_PORT_A(); \
			currentPortA = currentYSH = t; \
			SET(VIA_port_a, currentPortA); \
		} \
		else \
		{ \
			SET(VIA_port_b, 0x00); \
			currentYSH = currentPortA; \
		} \
	} \
	DELAY_YSH(); \
} while (0)

// sets 16 bit values divided by current scale!
// actually: due to possible display scaling, input values can now be larger 16bit
// if no value set, this "at least" switches the MUX off
#define SET_XSH16(v) \
  do { \
	int16_t t = (uint8_t)((v)/(currentScale+SCALE_STRENGTH_DIF));  \
	if (t != currentPortA) \
	{ \
		SET(VIA_port_b, 0x01); \
		DELAY_PORT_B_BEFORE_PORT_A(); \
		currentPortA = t; \
		SET(VIA_port_a, currentPortA); \
	}  \
	else \
  { \
		SET(VIA_port_b, 0x01); \
  } \
	DELAY_XSH(); \
} while (0)

// 4 cycles per loop?
// 1 cycle = 1 nano second
#define CYCLE_PER_LOOP 4
#define WAIT_CYCLE_NANO(n) do{ \
    uint32_t l = ((n)>>2)-((n)>>5)+((n)>>7)-((n)>>9); /* divided by 4 */  \
    if (!l) {l=1;} \
    else { l+=(n)>>7; \
    }asm volatile( "0:" "SUBS %[count], #2;" "SUBS %[count], #-1;" "BNE 0b;" :[count]"+r"(l) ); \
    } while(0)

#define GET_SYSTEM_TIMER_LO *(bcm2835_st + BCM2835_ST_CLO/4)
#define SET_TIMER_MARK do{timerMark = (volatile uint32_t)*(bcm2835_st + BCM2835_ST_CLO/4);} while(0)

// 32 bit timer, perhaps reset each timing?
// adjust overhead accordingly
// 1 tick = 0,001 us
// ticks * 1000 = 1000 us
#define WAIT_MICRO_TIME(utime) do{ \
       __sync_synchronize(); \
       uint32_t tstart = GET_SYSTEM_TIMER_LO; \
       uint32_t tend = tstart +(utime); \
       __sync_synchronize(); \
       while (GET_SYSTEM_TIMER_LO < tend);} while (0)

#define WAIT_MICRO_MARK_DELTA(utime)  do{ \\
       uint32_t tend = timerMark +(utime); \
       __sync_synchronize(); \
       while (GET_SYSTEM_TIMER_LO < tend);} while (0)


#define OFFSET_CYCLE_OVERHEAD 17// # of cycles "wasted" by function calling etc
#define OFFSET_TIMER_OVERHEAD 24// # of cycles "wasted" by function calling etc

// max 4000 micro
#define delayMicro(u) delayNano(((u)*1000))

// max 4 milli
#define delayMilli(u) delayNano(((u)*10000000))
#endif /* _VECTREX_INTERFACE_H */
